@@ -11,8 +11,7 @@ from io import BytesIO
 from itertools import zip_longest
 from pathlib import Path
 from shutil import copyfileobj
-from zlib import decompress
-from zlib import error as ZlibError
+from typing import NamedTuple, Optional
 
 from PIL import Image, ImageCms
 from PIL.TiffTags import TAGS_V2 as TIFF_TAGS
@@ -22,6 +21,7 @@ from pikepdf import (
     Dictionary,
     Name,
     Object,
+    Pdf,
     PdfError,
     Stream,
     StreamDecodeLevel,
@@ -31,24 +31,32 @@ from pikepdf import (
 
 
 class DependencyError(Exception):
-    pass
+    "A third party dependency is needed to extract images of this type."
 
 
 class UnsupportedImageTypeError(Exception):
-    pass
+    "This image is formatted in a way pikepdf does not supported."
 
 
 class NotExtractableError(Exception):
-    pass
+    "Indicates that an image cannot be directly extracted."
+
+
+class HifiPrintImageNotTranscodableError(NotExtractableError):
+    "Image contains high fidelity printing information and cannot be extracted."
+
+
+class InvalidPdfImageError(Exception):
+    "This image is not valid according to the PDF 1.7 specification."
 
 
 def array_str(value):
-    """Simplify pikepdf objects to array of str. Keep Streams intact."""
+    """Simplify pikepdf objects to array of str. Keep Streams and dictionaries intact."""
 
     def _array_str(item):
         if isinstance(item, (list, Array)):
             return [_array_str(subitem) for subitem in item]
-        elif isinstance(item, (Stream, bytes, int)):
+        elif isinstance(item, (Stream, Dictionary, bytes, int)):
             return item
         elif isinstance(item, (Name, str)):
             return str(item)
@@ -69,7 +77,7 @@ def dict_or_array_dict(value):
     if isinstance(value, Dictionary):
         return [value.as_dict()]
     if isinstance(value, Array):
-        return [v.as_list() for v in value]
+        return [v for v in value.as_list()]
     raise NotImplementedError(value)
 
 
@@ -83,53 +91,67 @@ def metadata_from_obj(obj, name, type_, default):
     raise NotImplementedError('Metadata access for ' + name)
 
 
+class PaletteData(NamedTuple):
+    """Returns the color space and binary representation of the palette.
+
+    ``base_colorspace`` is typically ``"RGB"`` or ``"L"`` (for grayscale).
+
+    ``palette`` is typically 256 or 256*3=768 bytes, for grayscale and RGB color
+    respectively, with each unit/triplet being the grayscale/RGB triplet values.
+    """
+
+    base_colorspace: str
+    palette: bytes
+
+
 class PdfImageBase(ABC):
 
     SIMPLE_COLORSPACES = {'/DeviceRGB', '/DeviceGray', '/CalRGB', '/CalGray'}
     MAIN_COLORSPACES = SIMPLE_COLORSPACES | {'/DeviceCMYK', '/CalCMYK', '/ICCBased'}
+    PRINT_COLORSPACES = {'/Separation', '/DeviceN'}
 
     @abstractmethod
     def _metadata(self, name, type_, default):
-        pass
+        """Get metadata for this image type."""
 
     @property
     def width(self):
-        """Width of the image data in pixels"""
+        """Width of the image data in pixels."""
         return self._metadata('Width', int, None)
 
     @property
     def height(self):
-        """Height of the image data in pixels"""
+        """Height of the image data in pixels."""
         return self._metadata('Height', int, None)
 
     @property
     def image_mask(self):
-        """``True`` if this is an image mask"""
+        """``True`` if this is an image mask."""
         return self._metadata('ImageMask', bool, False)
 
     @property
     def _bpc(self):
-        """Bits per component for this image (low-level)"""
+        """Bits per component for this image (low-level)."""
         return self._metadata('BitsPerComponent', int, None)
 
     @property
     def _colorspaces(self):
-        """Colorspace (low-level)"""
+        """Colorspace (low-level)."""
         return self._metadata('ColorSpace', array_str, [])
 
     @property
     def filters(self):
-        """List of names of the filters that we applied to encode this image"""
+        """List of names of the filters that we applied to encode this image."""
         return self._metadata('Filter', array_str, [])
 
     @property
     def decode_parms(self):
-        """List of the /DecodeParms, arguments to filters"""
+        """List of the /DecodeParms, arguments to filters."""
         return self._metadata('DecodeParms', dict_or_array_dict, [])
 
     @property
-    def colorspace(self):
-        """PDF name of the colorspace that best describes this image"""
+    def colorspace(self) -> Optional[str]:
+        """PDF name of the colorspace that best describes this image."""
         if self.image_mask:
             return None  # Undefined for image masks
         if self._colorspaces:
@@ -139,39 +161,81 @@ class PdfImageBase(ABC):
                 subspace = self._colorspaces[1]
                 if isinstance(subspace, str) and subspace in self.MAIN_COLORSPACES:
                     return subspace
-                if isinstance(subspace, list) and subspace[0] == '/ICCBased':
+                if isinstance(subspace, list) and subspace[0] in (
+                    '/ICCBased',
+                    '/DeviceN',
+                ):
                     return subspace[0]
+            if self._colorspaces[0] == '/DeviceN':
+                return '/DeviceN'
+
         raise NotImplementedError(
             "not sure how to get colorspace: " + repr(self._colorspaces)
         )
 
     @property
     def bits_per_component(self):
-        """Bits per component of this image"""
+        """Bits per component of this image."""
         if self._bpc is None:
             return 1 if self.image_mask else 8
         return self._bpc
 
     @property
     @abstractmethod
-    def is_inline(self):
-        pass
+    def is_inline(self) -> bool:
+        """Is this an inline image?"""
 
     @property
     @abstractmethod
-    def icc(self):
-        pass
+    def icc(self) -> Image.Image:
+        """Returns ICC profile for this image if one is defined."""
 
     @property
     def indexed(self):
-        """``True`` if the image has a defined color palette"""
+        """``True`` if the image has a defined color palette."""
         return '/Indexed' in self._colorspaces
 
+    def _colorspace_has_name(self, name):
+        try:
+            cs = self._colorspaces
+            if cs[0] == '/Indexed' and cs[1][0] == name:
+                return True
+            if cs[0] == name:
+                return True
+        except (IndexError, AttributeError, KeyError):
+            pass
+        return False
+
+    @property
+    def is_device_n(self):
+        """``True`` if image has a /DeviceN (complex printing) colorspace."""
+        return self._colorspace_has_name('/DeviceN')
+
+    @property
+    def is_separation(self):
+        """``True`` if image has a /DeviceN (complex printing) colorspace."""
+        return self._colorspace_has_name('/Separation')
+
     @property
     def size(self):
-        """Size of image as (width, height)"""
+        """Size of image as (width, height)."""
         return self.width, self.height
 
+    def _approx_mode_from_icc(self):
+        if self.indexed:
+            icc_profile = self._colorspaces[1][1]
+        else:
+            icc_profile = self._colorspaces[1]
+        icc_profile_nchannels = int(icc_profile['/N'])
+
+        if icc_profile_nchannels == 1:
+            return 'L'
+
+        # Multiple channels, need to open the profile and look
+        mode_from_xcolor_space = {'RGB ': 'RGB', 'CMYK': 'CMYK'}
+        xcolor_space = self.icc.profile.xcolor_space
+        return mode_from_xcolor_space.get(xcolor_space, '')
+
     @property
     def mode(self):
         """``PIL.Image.mode`` equivalent for this image, where possible
@@ -181,7 +245,11 @@ class PdfImageBase(ABC):
         """
 
         m = ''
-        if self.indexed:
+        if self.is_device_n:
+            m = 'DeviceN'
+        elif self.is_separation:
+            m = 'Separation'
+        elif self.indexed:
             m = 'P'
         elif self.bits_per_component == 1:
             m = '1'
@@ -194,18 +262,15 @@ class PdfImageBase(ABC):
                 m = 'CMYK'
             elif self.colorspace == '/ICCBased':
                 try:
-                    icc_profile = self._colorspaces[1]
-                    icc_profile_nchannels = int(icc_profile['/N'])
-                    if icc_profile_nchannels == 1:
-                        m = 'L'
-                    elif icc_profile_nchannels == 3:
-                        m = 'RGB'
-                    elif icc_profile_nchannels == 4:
-                        m = 'CMYK'
-                except (ValueError, TypeError):
-                    pass
+                    m = self._approx_mode_from_icc()
+                except (ValueError, TypeError) as e:
+                    raise NotImplementedError(
+                        "Not sure how to handle PDF image of this type"
+                    ) from e
         if m == '':
-            raise NotImplementedError("Not sure how to handle PDF image of this type")
+            raise NotImplementedError(
+                "Not sure how to handle PDF image of this type"
+            ) from None
         return m
 
     @property
@@ -223,12 +288,8 @@ class PdfImageBase(ABC):
         return list(zip_longest(self.filters, self.decode_parms, fillvalue={}))
 
     @property
-    def palette(self):
-        """Retrieves the color palette for this image
-
-        Returns:
-            tuple (base_colorspace: str, palette: bytes)
-        """
+    def palette(self) -> Optional[PaletteData]:
+        """Retrieves the color palette for this image if applicable."""
 
         if not self.indexed:
             return None
@@ -236,64 +297,78 @@ class PdfImageBase(ABC):
             _idx, base, _hival, lookup = self._colorspaces
         except ValueError as e:
             raise ValueError('Not sure how to interpret this palette') from e
-        iccobj = None
-        if self.icc:
-            iccobj = base[1]
+        if self.icc or self.is_device_n or self.is_separation:
             base = str(base[0])
         else:
             base = str(base)
-        _hival = int(_hival)
         lookup = bytes(lookup)
-        if not base in self.MAIN_COLORSPACES:
-            raise NotImplementedError("not sure how to interpret this palette")
+        if base not in self.MAIN_COLORSPACES and base not in self.PRINT_COLORSPACES:
+            raise NotImplementedError(f"not sure how to interpret this palette: {base}")
         if base == '/DeviceRGB':
             base = 'RGB'
         elif base == '/DeviceGray':
             base = 'L'
         elif base == '/DeviceCMYK':
             base = 'CMYK'
+        elif base == '/DeviceN':
+            base = 'DeviceN'
+        elif base == '/Separation':
+            base = 'Separation'
         elif base == '/ICCBased':
-            if iccobj['/N'] == 3:
-                base = 'RGB'
-            elif iccobj['/N'] == 4:
-                base = 'CMYK'
-            elif iccobj['/N'] == 1:
-                base = 'L'
-        return base, lookup
+            base = self._approx_mode_from_icc()
+        return PaletteData(base, lookup)
 
     @abstractmethod
-    def as_pil_image(self):
-        pass
+    def as_pil_image(self) -> Image.Image:
+        """Convert this PDF image to a Python PIL (Pillow) image."""
 
     @staticmethod
-    def _unstack_compression(buffer, filters):
-        """Remove stacked compression where it appears.
-
-        Stacked compression means when an image is set to:
-            ``[/FlateDecode /DCTDecode]``
-        for example.
-
-        Only Flate can be stripped off the front currently.
+    def _remove_simple_filters(obj, filters):
+        """Remove simple lossless compression where it appears.
 
         Args:
-            buffer (pikepdf._qpdf.Buffer): the compressed image data
+            obj (pikepdf.Stream): the compressed object
             filters (list of str): all files on the data
         """
-        data = memoryview(buffer)
-        while len(filters) > 1 and filters[0] == '/FlateDecode':
-            try:
-                data = decompress(data)
-            except ZlibError as e:
-                raise UnsupportedImageTypeError() from e
-            filters = filters[1:]
-        return data, filters
+
+        COMPLEX_FILTERS = {
+            '/DCTDecode',
+            '/JPXDecode',
+            '/JBIG2Decode',
+            '/CCITTFaxDecode',
+        }
+
+        idx = [n for n, item in enumerate(filters) if item in COMPLEX_FILTERS]
+        if idx:
+            if len(idx) > 1:
+                raise NotImplementedError(
+                    f"Object {obj.objgen} has compound complex filters: {filters}. "
+                    "We cannot decompress this."
+                )
+            simple_filters = filters[: idx[0]]
+            complex_filters = filters[idx[0] :]
+        else:
+            simple_filters = filters
+            complex_filters = []
+
+        if not simple_filters:
+            return obj.read_raw_bytes(), complex_filters
+
+        original_filters = obj.Filter
+        try:
+            obj.Filter = Array([Name(s) for s in simple_filters])
+            data = obj.read_bytes(StreamDecodeLevel.specialized)
+        finally:
+            obj.Filter = original_filters
+
+        return data, complex_filters
 
 
 class PdfImage(PdfImageBase):
     """Support class to provide a consistent API for manipulating PDF images
 
-    The data structure for images inside PDFs is irregular and flexible,
-    making it difficult to work with without introducing errors for less
+    The data structure for images inside PDFs is irregular and complex,
+    making it difficult to use without introducing errors for less
     typical cases. This class addresses these difficulties by providing a
     regular, Pythonic API similar in spirit (and convertible to) the Python
     Pillow imaging library.
@@ -322,6 +397,8 @@ class PdfImage(PdfImageBase):
         self._icc = None
 
     def __eq__(self, other):
+        if not isinstance(other, PdfImageBase):
+            return NotImplemented
         return self.obj == other.obj
 
     @classmethod
@@ -357,7 +434,7 @@ class PdfImage(PdfImageBase):
 
     @property
     def is_inline(self):
-        """``False`` for image XObject"""
+        """``False`` for image XObject."""
         return False
 
     @property
@@ -384,10 +461,17 @@ class PdfImage(PdfImageBase):
             iccstream = self._iccstream
             iccbuffer = iccstream.get_stream_buffer()
             iccbytesio = BytesIO(iccbuffer)
-            self._icc = ImageCms.ImageCmsProfile(iccbytesio)
+            try:
+                self._icc = ImageCms.ImageCmsProfile(iccbytesio)
+            except OSError as e:
+                if str(e) == 'cannot open profile from string':
+                    # ICC profile is corrupt
+                    raise UnsupportedImageTypeError(
+                        "ICC profile corrupt or not readable"
+                    ) from e
         return self._icc
 
-    def _extract_direct(self, *, stream):
+    def _extract_direct(self, *, stream) -> str:
         """Attempt to extract the image directly to a usable image file
 
         If there is no way to extract the image without decompressing or
@@ -415,9 +499,7 @@ class PdfImage(PdfImageBase):
             ct = self.filter_decodeparms[0][1].get('/ColorTransform', DEFAULT_CT_CMYK)
             return self.mode == 'CMYK' and ct == DEFAULT_CT_CMYK
 
-        data, filters = self._unstack_compression(
-            self.obj.get_raw_stream_buffer(), self.filters
-        )
+        data, filters = self._remove_simple_filters(self.obj, self.filters)
 
         if filters == ['/CCITTFaxDecode']:
             if self.colorspace == '/ICCBased':
@@ -435,58 +517,99 @@ class PdfImage(PdfImageBase):
 
         raise NotExtractableError()
 
-    def _extract_transcoded(self):
+    def _extract_transcoded(self) -> Image.Image:
+        # Reminder Pillow palette byte order unintentionally changed in 8.3.0
+        # https://github.com/python-pillow/Pillow/issues/5595
+        # 8.2.0: all aligned by channel (very nonstandard)
+        # 8.3.0: all channels for one color followed by the next color (e.g. RGBRGBRGB)
+
         im = None
+        if self.mode in {'DeviceN', 'Separation'}:
+            raise HifiPrintImageNotTranscodableError()
+
         if self.mode == 'RGB' and self.bits_per_component == 8:
             # No point in accessing the buffer here, size qpdf decodes to 3-byte
             # RGB and Pillow needs RGBX for raw access
-            data = self.read_bytes()
-            im = Image.frombytes('RGB', self.size, data)
+            im = Image.frombuffer(
+                'RGB', self.size, self.read_bytes(), 'raw', 'RGB', 0, 1
+            )
+        elif self.mode == 'CMYK' and self.bits_per_component == 8:
+            im = Image.frombuffer(
+                'CMYK', self.size, self.get_stream_buffer(), 'raw', 'CMYK', 0, 1
+            )
         elif self.mode in ('L', 'P') and self.bits_per_component == 8:
             buffer = self.get_stream_buffer()
             stride = 0  # tell Pillow to calculate stride from line width
             ystep = 1  # image is top to bottom in memory
             im = Image.frombuffer('L', self.size, buffer, "raw", 'L', stride, ystep)
-            if self.mode == 'P':
+            if self.mode == 'P' and self.palette is not None:
                 base_mode, palette = self.palette
-                if base_mode in ('RGB', 'L'):
+                if base_mode == 'RGB':
                     im.putpalette(palette, rawmode=base_mode)
+                elif base_mode == 'L':
+                    # Pillow does not fully support palettes with rawmode='L'.
+                    # Convert to RGB palette.
+                    gray_palette = palette
+                    palette = b''
+                    for entry in gray_palette:
+                        palette += bytes([entry]) * 3
+                    im.putpalette(palette, rawmode='RGB')
+                elif base_mode == 'CMYK':
+                    # Pillow does not support CMYK with palettes; convert manually
+                    with memoryview(buffer) as mv:
+                        output = bytearray(4 * len(mv))
+                        for n, pal_idx in enumerate(mv):
+                            output[4 * n : 4 * (n + 1)] = palette[
+                                4 * pal_idx : 4 * (pal_idx + 1)
+                            ]
+                    im = Image.frombuffer(
+                        'CMYK', self.size, data=output, decoder_name='raw'
+                    )
                 else:
-                    raise NotImplementedError('palette with ' + base_mode)
+                    raise NotImplementedError(f'palette with {base_mode}')
         elif self.bits_per_component == 1:
             if self.filters and self.filters[0] == '/JBIG2Decode':
                 if not jbig2.jbig2dec_available():
-                    raise DependencyError("jbig2dec - not installed")
+                    raise DependencyError(
+                        "jbig2dec - not installed or installed version is too old "
+                        "(older than version 0.15)"
+                    )
                 jbig2_globals_obj = self.filter_decodeparms[0][1].get('/JBIG2Globals')
                 im = jbig2.extract_jbig2(self.obj, jbig2_globals_obj)
             else:
                 data = self.read_bytes()
                 im = Image.frombytes('1', self.size, data)
         else:
-            raise UnsupportedImageTypeError(repr(self))
+            raise UnsupportedImageTypeError(repr(self) + ", " + repr(self.obj))
 
-        if self.mode == 'P' and self.bits_per_component == 1:
+        if (
+            self.mode == 'P'
+            and self.bits_per_component == 1
+            and self.palette is not None
+        ):
             # Fix paletted 1-bit images
             base_mode, palette = self.palette
             if base_mode == 'RGB' and palette != b'\x00\x00\x00\xff\xff\xff':
                 im = im.convert('P')
                 im.putpalette(palette, rawmode=base_mode)
                 gp = im.getpalette()
-                gp[765:768] = gp[3:6]  # work around Pillow bug
-                im.putpalette(gp)
+                if gp:
+                    gp[765:768] = gp[3:6]  # work around Pillow bug
+                    im.putpalette(gp)
             elif base_mode == 'L' and palette != b'\x00\xff':
                 im = im.convert('P')
                 im.putpalette(palette, rawmode=base_mode)
                 gp = im.getpalette()
-                gp[255] = gp[1]  # work around Pillow bug
-                im.putpalette(gp)
+                if gp:
+                    gp[255] = gp[1]  # work around Pillow bug
+                    im.putpalette(gp)
 
         if self.colorspace == '/ICCBased':
             im.info['icc_profile'] = self.icc.tobytes()
 
         return im
 
-    def _extract_to_stream(self, *, stream):
+    def _extract_to_stream(self, *, stream) -> str:
         """Attempt to extract the image to a stream
 
         If possible, the compressed data is extracted and inserted into
@@ -498,7 +621,7 @@ class PdfImage(PdfImageBase):
             stream: Writable stream to write data to
 
         Returns:
-            str: The file format extension
+            The file format extension.
         """
 
         try:
@@ -506,19 +629,26 @@ class PdfImage(PdfImageBase):
         except NotExtractableError:
             pass
 
+        im = None
         try:
             im = self._extract_transcoded()
-            if im:
+            if im.mode == 'CMYK':
+                im.save(stream, format='tiff', compression='tiff_adobe_deflate')
+                return '.tiff'
+            elif im:
                 im.save(stream, format='png')
                 return '.png'
         except PdfError as e:
-            if 'getStreamData called on unfilterable stream' in str(e):
+            if 'called on unfilterable stream' in str(e):
                 raise UnsupportedImageTypeError(repr(self)) from e
             raise
+        finally:
+            if im:
+                im.close()
 
         raise UnsupportedImageTypeError(repr(self))
 
-    def extract_to(self, *, stream=None, fileprefix=''):
+    def extract_to(self, *, stream=None, fileprefix='') -> str:
         """Attempt to extract the image directly to a usable image file
 
         If possible, the compressed data is extracted and inserted into
@@ -531,6 +661,8 @@ class PdfImage(PdfImageBase):
         When saving the image to a file, use a temporary filename, and then
         rename the file to its final name based on the returned file extension.
 
+        Images might be saved as any of .png, .jpg, or .tiff.
+
         Examples:
 
             >>> im.extract_to(stream=bytes_io)
@@ -548,9 +680,6 @@ class PdfImage(PdfImageBase):
             If *fileprefix* was provided, then the fileprefix with the
             appropriate extension. If no *fileprefix*, then an extension
             indicating the file type.
-
-        Return type:
-            str
         """
 
         if bool(stream) == bool(fileprefix):
@@ -567,18 +696,17 @@ class PdfImage(PdfImageBase):
         return str(filepath)
 
     def read_bytes(self, decode_level=StreamDecodeLevel.specialized):
-        """Decompress this image and return it as unencoded bytes"""
+        """Decompress this image and return it as unencoded bytes."""
         return self.obj.read_bytes(decode_level=decode_level)
 
     def get_stream_buffer(self, decode_level=StreamDecodeLevel.specialized):
-        """Access this image with the buffer protocol"""
+        """Access this image with the buffer protocol."""
         return self.obj.get_stream_buffer(decode_level=decode_level)
 
-    def as_pil_image(self):
-        """Extract the image as a Pillow Image, using decompression as necessary
+    def as_pil_image(self) -> Image.Image:
+        """Extract the image as a Pillow Image, using decompression as necessary.
 
-        Returns:
-            PIL.Image.Image
+        Caller must close the image.
         """
         try:
             bio = BytesIO()
@@ -595,7 +723,7 @@ class PdfImage(PdfImageBase):
         return im
 
     def _generate_ccitt_header(self, data, icc=None):
-        """Construct a CCITT G3 or G4 header from the PDF metadata"""
+        """Construct a CCITT G3 or G4 header from the PDF metadata."""
         # https://stackoverflow.com/questions/2641770/
         # https://www.itu.int/itudoc/itu-t/com16/tiff-fx/docs/tiff6.pdf
 
@@ -613,7 +741,7 @@ class PdfImage(PdfImageBase):
         elif k > 0:
             ccitt_group = 3  # Group 3 2-D
         else:
-            ccitt_group = 2  # CCITT 1-D
+            ccitt_group = 2  # Group 3 1-D
         black_is_one = self.decode_parms[0].get("/BlackIs1", False)
         # TIFF spec says:
         # use 0 for white_is_zero (=> black is 1)
@@ -677,32 +805,42 @@ class PdfImage(PdfImageBase):
 
         return tiff_header
 
-    def show(self):
-        """Show the image however PIL wants to"""
+    def show(self):  # pragma: no cover
+        """Show the image however PIL wants to."""
         self.as_pil_image().show()
 
     def __repr__(self):
-        return '<pikepdf.PdfImage image mode={} size={}x{} at {}>'.format(
-            self.mode, self.width, self.height, hex(id(self))
+        return (
+            f'<pikepdf.PdfImage image mode={self.mode} '
+            f'size={self.width}x{self.height} at {hex(id(self))}>'
         )
 
     def _repr_png_(self):
-        """Display hook for IPython/Jupyter"""
+        """Display hook for IPython/Jupyter."""
         b = BytesIO()
-        im = self.as_pil_image()
-        im.save(b, 'PNG')
-        return b.getvalue()
+        with self.as_pil_image() as im:
+            im.save(b, 'PNG')
+            return b.getvalue()
 
 
 class PdfJpxImage(PdfImage):
+    """Support class for JPEG 2000 images. Implements the same API as :class:`PdfImage`."""
+
     def __init__(self, obj):
         super().__init__(obj)
-        self.pil = self.as_pil_image()
+        self._jpxpil = self.as_pil_image()
+
+    def __eq__(self, other):
+        if not isinstance(other, PdfImageBase):
+            return NotImplemented
+        return (
+            self.obj == other.obj
+            and isinstance(other, PdfJpxImage)
+            and self._jpxpil == other._jpxpil
+        )
 
     def _extract_direct(self, *, stream):
-        data, filters = self._unstack_compression(
-            self.obj.get_raw_stream_buffer(), self.filters
-        )
+        data, filters = self._remove_simple_filters(self.obj, self.filters)
         if filters != ['/JPXDecode']:
             raise UnsupportedImageTypeError(self.filters)
         stream.write(data)
@@ -715,9 +853,9 @@ class PdfJpxImage(PdfImage):
         super_colorspaces = super()._colorspaces
         if super_colorspaces:
             return super_colorspaces
-        if self.pil.mode == 'L':
+        if self._jpxpil.mode == 'L':
             return ['/DeviceGray']
-        elif self.pil.mode == 'RGB':
+        elif self._jpxpil.mode == 'RGB':
             return ['/DeviceRGB']
         raise NotImplementedError('Complex JP2 colorspace')
 
@@ -737,13 +875,14 @@ class PdfJpxImage(PdfImage):
         return False
 
     def __repr__(self):
-        return '<pikepdf.PdfJpxImage JPEG2000 image mode={} size={}x{} at {}>'.format(
-            self.mode, self.width, self.height, hex(id(self))
+        return (
+            f'<pikepdf.PdfJpxImage JPEG2000 image mode={self.mode} '
+            f'size={self.width}x{self.height} at {hex(id(self))}>'
         )
 
 
 class PdfInlineImage(PdfImageBase):
-    """Support class for PDF inline images"""
+    """Support class for PDF inline images. Implements the same API as :class:`PdfImage`."""
 
     # Inline images can contain abbreviations that we write automatically
     ABBREVS = {
@@ -765,8 +904,9 @@ class PdfInlineImage(PdfImageBase):
         b'/CCF': b'/CCITTFaxDecode',
         b'/DCT': b'/DCTDecode',
     }
+    REVERSE_ABBREVS = {v: k for k, v in ABBREVS.items()}
 
-    def __init__(self, *, image_data, image_object: tuple):
+    def __init__(self, *, image_data: Object, image_object: tuple):
         """
         Args:
             image_data: data stream for image, extracted from content stream
@@ -781,29 +921,34 @@ class PdfInlineImage(PdfImageBase):
         self._data = image_data
         self._image_object = image_object
 
-        reparse = b' '.join(self._unparse_obj(obj) for obj in image_object)
+        reparse = b' '.join(
+            self._unparse_obj(obj, remap_names=self.ABBREVS) for obj in image_object
+        )
         try:
             reparsed_obj = Object.parse(b'<< ' + reparse + b' >>')
         except PdfError as e:
             raise PdfError("parsing inline " + reparse.decode('unicode_escape')) from e
         self.obj = reparsed_obj
-        self.pil = None
 
     def __eq__(self, other):
-        if self.obj == other.obj and (
-            self._data._inline_image_raw_bytes()
-            == other._data._inline_image_raw_bytes()
-        ):
-            return True
-        return False
+        if not isinstance(other, PdfImageBase):
+            return NotImplemented
+        return (
+            self.obj == other.obj
+            and isinstance(other, PdfInlineImage)
+            and (
+                self._data._inline_image_raw_bytes()
+                == other._data._inline_image_raw_bytes()
+            )
+        )
 
     @classmethod
-    def _unparse_obj(cls, obj):
+    def _unparse_obj(cls, obj, remap_names):
         if isinstance(obj, Object):
             if isinstance(obj, Name):
                 name = obj.unparse(resolved=True)
                 assert isinstance(name, bytes)
-                return cls.ABBREVS.get(name, name)
+                return remap_names.get(name, name)
             else:
                 return obj.unparse(resolved=True)
         elif isinstance(obj, bool):
@@ -817,47 +962,67 @@ class PdfInlineImage(PdfImageBase):
         return metadata_from_obj(self.obj, name, type_, default)
 
     def unparse(self):
-        tokens = []
-        tokens.append(b'BI\n')
-        metadata = []
-        for metadata_obj in self._image_object:
-            unparsed = self._unparse_obj(metadata_obj)
-            assert isinstance(unparsed, bytes)
-            metadata.append(unparsed)
-        tokens.append(b' '.join(metadata))
-        tokens.append(b'\nID\n')
-        tokens.append(self._data._inline_image_raw_bytes())
-        tokens.append(b'EI')
-        return b''.join(tokens)
+        def metadata_tokens():
+            for metadata_obj in self._image_object:
+                unparsed = self._unparse_obj(
+                    metadata_obj, remap_names=self.REVERSE_ABBREVS
+                )
+                assert isinstance(unparsed, bytes)
+                yield unparsed
+
+        def inline_image_tokens():
+            yield b'BI\n'
+            yield b' '.join(m for m in metadata_tokens())
+            yield b'\nID\n'
+            yield self._data._inline_image_raw_bytes()
+            yield b'EI'
+
+        return b''.join(inline_image_tokens())
 
     @property
     def is_inline(self):
         return True
 
     @property
-    def icc(self):
-        raise NotImplementedError("Inline images with ICC profiles are not supported")
-
-    def __repr__(self):
-        mode = '?'
-        try:
-            mode = self.mode
-        except Exception:
-            pass
-        return '<pikepdf.PdfInlineImage image mode={} size={}x{} at {}>'.format(
-            mode, self.width, self.height, hex(id(self))
+    def icc(self):  # pragma: no cover
+        raise InvalidPdfImageError(
+            "Inline images with ICC profiles are not supported in the PDF specification"
         )
 
+    def __repr__(self):
+        try:
+            mode = self.mode
+        except NotImplementedError:
+            mode = '?'
+        return (
+            f'<pikepdf.PdfInlineImage image mode={mode} '
+            f'size={self.width}x{self.height} at {hex(id(self))}>'
+        )
+
+    def _convert_to_pdfimage(self):
+        # Construct a temporary PDF that holds this inline image, and...
+        tmppdf = Pdf.new()
+        tmppdf.add_blank_page(page_size=(self.width, self.height))
+        tmppdf.pages[0].contents_add(
+            f'{self.width} 0 0 {self.height} 0 0 cm'.encode('ascii'), prepend=True
+        )
+        tmppdf.pages[0].contents_add(self.unparse())
+
+        # ...externalize it,
+        tmppdf.pages[0].externalize_inline_images()
+        raw_img = next(im for im in tmppdf.pages[0].images.values())
+
+        # ...then use the regular PdfImage API to extract it.
+        img = PdfImage(raw_img)
+        return img
+
     def as_pil_image(self):
-        if self.pil:
-            return self.pil
+        return self._convert_to_pdfimage().as_pil_image()
 
-        raise NotImplementedError('not yet')
-
-    def extract_to(
-        self, *, stream=None, fileprefix=''
-    ):  # pylint: disable=unused-argument
-        raise UnsupportedImageTypeError("inline images don't support extract")
+    def extract_to(self, *, stream=None, fileprefix=''):
+        return self._convert_to_pdfimage().extract_to(
+            stream=stream, fileprefix=fileprefix
+        )
 
     def read_bytes(self):
         raise NotImplementedError("qpdf returns compressed")
